package java.net;

import java.security.AccessController;
import java.security.PrivilegedAction;
import sun.net.ConnectionResetException;
import java.io.OutputStream;
import java.io.InputStream;
import sun.net.NetHooks;
import java.util.Objects;
import java.util.Collections;
import java.util.HashSet;
import java.io.IOException;
import sun.net.ResourceManager;
import java.io.FileDescriptor;
import java.util.Set;
import sun.net.ExtendedSocketOptions;

abstract class AbstractPlainSocketImpl extends SocketImpl
{
    int timeout;
    private int trafficClass;
    private boolean shut_rd;
    private boolean shut_wr;
    private SocketInputStream socketInputStream;
    private SocketOutputStream socketOutputStream;
    protected int fdUseCount;
    protected final Object fdLock;
    protected boolean closePending;
    private int CONNECTION_NOT_RESET;
    private int CONNECTION_RESET_PENDING;
    private int CONNECTION_RESET;
    private int resetState;
    private final Object resetLock;
    protected boolean stream;
    final boolean isServer;
    static final ExtendedSocketOptions extendedOptions;
    private static final Set<SocketOption<?>> clientSocketOptions;
    private static final Set<SocketOption<?>> serverSocketOptions;
    public static final int SHUT_RD = 0;
    public static final int SHUT_WR = 1;
    
    AbstractPlainSocketImpl(boolean isServer) {
        this.shut_rd = false;
        this.shut_wr = false;
        this.socketInputStream = null;
        this.socketOutputStream = null;
        this.fdUseCount = 0;
        this.fdLock = new Object();
        this.closePending = false;
        this.CONNECTION_NOT_RESET = 0;
        this.CONNECTION_RESET_PENDING = 1;
        this.CONNECTION_RESET = 2;
        this.resetLock = new Object();
        this.isServer = isServer;
    }
    
    @Override
    protected synchronized void create(boolean stream) throws IOException {
        if (!(this.stream = stream)) {
            ResourceManager.beforeUdpCreate();
            this.fd = new FileDescriptor();
            try {
                this.socketCreate(false);
            }
            catch (IOException ex) {
                ResourceManager.afterUdpClose();
                this.fd = null;
                throw ex;
            }
        }
        else {
            this.fd = new FileDescriptor();
            this.socketCreate(true);
        }
        if (this.socket != null) {
            this.socket.setCreated();
        }
        if (this.serverSocket != null) {
            this.serverSocket.setCreated();
        }
    }
    
    @Override
    protected void connect(String s, int port) throws UnknownHostException, IOException {
        boolean n = false;
        try {
            InetAddress byName = InetAddress.getByName(s);
            this.port = port;
            this.connectToAddress(this.address = byName, port, this.timeout);
            n = true;
        }
        finally {
            if (n == 0) {
                try {
                    this.close();
                }
                catch (IOException ex) {}
            }
        }
    }
    
    @Override
    protected void connect(InetAddress address, int port) throws IOException {
        this.port = port;
        this.address = address;
        try {
            this.connectToAddress(address, port, this.timeout);
            return;
        }
        catch (IOException ex) {
            this.close();
            throw ex;
        }
    }
    
    @Override
    protected void connect(SocketAddress socketAddress, int n) throws IOException {
        boolean n2 = false;
        try {
            if (socketAddress == null || !(socketAddress instanceof InetSocketAddress)) {
                throw new IllegalArgumentException("unsupported address type");
            }
            InetSocketAddress inetSocketAddress = (InetSocketAddress)socketAddress;
            if (inetSocketAddress.isUnresolved()) {
                throw new UnknownHostException(inetSocketAddress.getHostName());
            }
            this.port = inetSocketAddress.getPort();
            this.connectToAddress(this.address = inetSocketAddress.getAddress(), this.port, n);
            n2 = true;
        }
        finally {
            if (n2 == 0) {
                try {
                    this.close();
                }
                catch (IOException ex) {}
            }
        }
    }
    
    private void connectToAddress(InetAddress inetAddress, int n, int n2) throws IOException {
        if (inetAddress.isAnyLocalAddress()) {
            this.doConnect(InetAddress.getLocalHost(), n, n2);
        }
        else {
            this.doConnect(inetAddress, n, n2);
        }
    }
    
    @Override
    public void setOption(int i, Object o) throws SocketException {
        if (this.isClosedOrPending()) {
            throw new SocketException("Socket Closed");
        }
        boolean n = true;
        switch (i) {
            case 128: {
                if (o == null || (!(o instanceof Integer) && !(o instanceof Boolean))) {
                    throw new SocketException("Bad parameter for option");
                }
                if ((o instanceof Boolean)) {
                    n = false;
                    break;
                }
                break;
            }
            case 4102: {
                if (o == null || !(o instanceof Integer)) {
                    throw new SocketException("Bad parameter for SO_TIMEOUT");
                }
                int intValue = (int)o;
                if (intValue < 0) {
                    throw new IllegalArgumentException("timeout < 0");
                }
                this.timeout = intValue;
                break;
            }
            case 3: {
                if (o == null || !(o instanceof Integer)) {
                    throw new SocketException("bad argument for IP_TOS");
                }
                this.trafficClass = (int)o;
                break;
            }
            case 15: {
                throw new SocketException("Cannot re-bind socket");
            }
            case 1: {
                if (o == null || !(o instanceof Boolean)) {
                    throw new SocketException("bad parameter for TCP_NODELAY");
                }
                n = (boolean)o;
                break;
            }
            case 4097:
            case 4098: {
                if (o == null || !(o instanceof Integer) || (int)o <= 0) {
                    throw new SocketException("bad parameter for SO_SNDBUF or SO_RCVBUF");
                }
                break;
            }
            case 8: {
                if (o == null || !(o instanceof Boolean)) {
                    throw new SocketException("bad parameter for SO_KEEPALIVE");
                }
                n = (boolean)o;
                break;
            }
            case 4099: {
                if (o == null || !(o instanceof Boolean)) {
                    throw new SocketException("bad parameter for SO_OOBINLINE");
                }
                n = (boolean)o;
                break;
            }
            case 4: {
                if (o == null || !(o instanceof Boolean)) {
                    throw new SocketException("bad parameter for SO_REUSEADDR");
                }
                n = (boolean)o;
                break;
            }
            default: {
                throw new SocketException("unrecognized TCP option: " + i);
            }
        }
        this.socketSetOption(i, (boolean)n, o);
        return;
    }
    
    @Override
    public Object getOption(int n) throws SocketException {
        if (this.isClosedOrPending()) {
            throw new SocketException("Socket Closed");
        }
        if (n == 4102) {
            return new Integer(this.timeout);
        }
        switch (n) {
            case 1: {
                return Boolean.valueOf(this.socketGetOption(n, null) != -1);
            }
            case 4099: {
                return Boolean.valueOf(this.socketGetOption(n, null) != -1);
            }
            case 128: {
                int socketGetOption = this.socketGetOption(n, null);
                return (socketGetOption == -1) ? Boolean.FALSE : new Integer(socketGetOption);
            }
            case 4: {
                return Boolean.valueOf(this.socketGetOption(n, null) != -1);
            }
            case 15: {
                InetAddressContainer inetAddressContainer = new InetAddressContainer();
                this.socketGetOption(n, inetAddressContainer);
                return inetAddressContainer.addr;
            }
            case 4097:
            case 4098: {
                return new Integer(this.socketGetOption(n, null));
            }
            case 3: {
                try {
                    int socketGetOption2 = this.socketGetOption(n, null);
                    if (socketGetOption2 == -1) {
                        return Integer.valueOf(this.trafficClass);
                    }
                    return Integer.valueOf(socketGetOption2);
                }
                catch (SocketException ex) {
                    return Integer.valueOf(this.trafficClass);
                }
            }
            case 8: {
                return Boolean.valueOf(this.socketGetOption(n, null) != -1);
            }
            default: {
                return null;
            }
        }
    }
    
    private static Set<SocketOption<?>> clientSocketOptions() {
        HashSet<StandardSocketOptions> s = new HashSet();
        s.add(StandardSocketOptions.SO_KEEPALIVE);
        s.add(StandardSocketOptions.SO_SNDBUF);
        s.add(StandardSocketOptions.SO_RCVBUF);
        s.add(StandardSocketOptions.SO_REUSEADDR);
        s.add(StandardSocketOptions.SO_LINGER);
        s.add(StandardSocketOptions.IP_TOS);
        s.add(StandardSocketOptions.TCP_NODELAY);
        s.addAll(ExtendedSocketOptions.clientSocketOptions());
        return Collections.unmodifiableSet(s);
    }
    
    private static Set<SocketOption<?>> serverSocketOptions() {
        HashSet<StandardSocketOptions> s = new HashSet();
        s.add(StandardSocketOptions.SO_RCVBUF);
        s.add(StandardSocketOptions.SO_REUSEADDR);
        s.add(StandardSocketOptions.IP_TOS);
        s.addAll(ExtendedSocketOptions.serverSocketOptions());
        return Collections.unmodifiableSet(s);
    }
    
    protected Set<SocketOption<?>> supportedOptions() {
        if (this.isServer) {
            return AbstractPlainSocketImpl.serverSocketOptions;
        }
        return AbstractPlainSocketImpl.clientSocketOptions;
    }
    
    protected <T> void setOption(SocketOption<T> obj, T t) throws IOException {
        Objects.requireNonNull(obj);
        if (!this.supportedOptions().contains(obj)) {
            throw new UnsupportedOperationException("'" + obj + "' not supported");
        }
        if (!obj.type().isInstance(t)) {
            throw new IllegalArgumentException("Invalid value '" + t + "'");
        }
        if (this.isClosedOrPending()) {
            throw new SocketException("Socket closed");
        }
        if (obj == StandardSocketOptions.SO_KEEPALIVE) {
            this.setOption(8, t);
        }
        else if (obj == StandardSocketOptions.SO_SNDBUF) {
            if ((int)t < 0) {
                throw new IllegalArgumentException("Invalid send buffer size:" + t);
            }
            this.setOption(4097, t);
        }
        else if (obj == StandardSocketOptions.SO_RCVBUF) {
            if ((int)t < 0) {
                throw new IllegalArgumentException("Invalid recv buffer size:" + t);
            }
            this.setOption(4098, t);
        }
        else if (obj == StandardSocketOptions.SO_REUSEADDR) {
            this.setOption(4, t);
        }
        else if (obj == StandardSocketOptions.SO_LINGER) {
            if ((int)t < 0) {
                this.setOption(128, Boolean.valueOf(false));
            }
            else {
                this.setOption(128, t);
            }
        }
        else if (obj == StandardSocketOptions.IP_TOS) {
            int intValue = (int)t;
            if (intValue < 0 || intValue > 255) {
                throw new IllegalArgumentException("Invalid IP_TOS value: " + t);
            }
            this.setOption(3, t);
        }
        else if (obj == StandardSocketOptions.TCP_NODELAY) {
            this.setOption(1, t);
        }
        else if (AbstractPlainSocketImpl.extendedOptions.isOptionSupported(obj)) {
            AbstractPlainSocketImpl.extendedOptions.setOption(this.fd, obj, t);
        }
        else {
            throw new AssertionError("unknown option: " + obj);
        }
        return;
    }
    
    protected <T> T getOption(SocketOption<T> obj) throws IOException {
        Objects.requireNonNull(obj);
        if (!this.supportedOptions().contains(obj)) {
            throw new UnsupportedOperationException("'" + obj + "' not supported");
        }
        if (this.isClosedOrPending()) {
            throw new SocketException("Socket closed");
        }
        if (obj == StandardSocketOptions.SO_KEEPALIVE) {
            return this.getOption(8);
        }
        if (obj == StandardSocketOptions.SO_SNDBUF) {
            return this.getOption(4097);
        }
        if (obj == StandardSocketOptions.SO_RCVBUF) {
            return this.getOption(4098);
        }
        if (obj == StandardSocketOptions.SO_REUSEADDR) {
            return this.getOption(4);
        }
        if (obj == StandardSocketOptions.SO_LINGER) {
            Object o = this.getOption(128);
            if ((o instanceof Boolean)) {
                assert !(boolean)o;
                o = Integer.valueOf(-1);
            }
            return o;
        }
        else {
            if (obj == StandardSocketOptions.IP_TOS) {
                return this.getOption(3);
            }
            if (obj == StandardSocketOptions.TCP_NODELAY) {
                return this.getOption(1);
            }
            if (AbstractPlainSocketImpl.extendedOptions.isOptionSupported(obj)) {
                return AbstractPlainSocketImpl.extendedOptions.getOption(this.fd, obj);
            }
            throw new AssertionError("unknown option: " + obj);
        }
    }
    
    synchronized void doConnect(InetAddress address, int port, int n) throws IOException {
        synchronized (this.fdLock) {
            if (!this.closePending && (this.socket == null || !this.socket.isBound())) {
                NetHooks.beforeTcpConnect(this.fd, address, port);
            }
        }
        try {
            this.acquireFD();
            try {
                this.socketConnect(address, port, n);
                synchronized (this.fdLock) {
                    if (this.closePending) {
                        throw new SocketException("Socket closed");
                    }
                }
                if (this.socket != null) {
                    this.socket.setBound();
                    this.socket.setConnected();
                }
            }
            finally {
                this.releaseFD();
            }
        }
        catch (IOException ex) {
            this.close();
            throw ex;
        }
    }
    
    @Override
    protected synchronized void bind(InetAddress address, int port) throws IOException {
        synchronized (this.fdLock) {
            if (!this.closePending && (this.socket == null || !this.socket.isBound())) {
                NetHooks.beforeTcpBind(this.fd, address, port);
            }
        }
        this.socketBind(address, port);
        if (this.socket != null) {
            this.socket.setBound();
        }
        if (this.serverSocket != null) {
            this.serverSocket.setBound();
        }
    }
    
    @Override
    protected synchronized void listen(int n) throws IOException {
        this.socketListen(n);
    }
    
    @Override
    protected void accept(SocketImpl socketImpl) throws IOException {
        this.acquireFD();
        try {
            this.socketAccept(socketImpl);
        }
        finally {
            this.releaseFD();
        }
    }
    
    @Override
    protected synchronized InputStream getInputStream() throws IOException {
        synchronized (this.fdLock) {
            if (this.isClosedOrPending()) {
                throw new IOException("Socket Closed");
            }
            if (this.shut_rd) {
                throw new IOException("Socket input is shutdown");
            }
            if (this.socketInputStream == null) {
                this.socketInputStream = new SocketInputStream(this);
            }
        }
        return this.socketInputStream;
    }
    
    void setInputStream(SocketInputStream socketInputStream) {
        this.socketInputStream = socketInputStream;
    }
    
    @Override
    protected synchronized OutputStream getOutputStream() throws IOException {
        synchronized (this.fdLock) {
            if (this.isClosedOrPending()) {
                throw new IOException("Socket Closed");
            }
            if (this.shut_wr) {
                throw new IOException("Socket output is shutdown");
            }
            if (this.socketOutputStream == null) {
                this.socketOutputStream = new SocketOutputStream(this);
            }
        }
        return this.socketOutputStream;
    }
    
    void setFileDescriptor(FileDescriptor fd) {
        this.fd = fd;
    }
    
    void setAddress(InetAddress address) {
        this.address = address;
    }
    
    void setPort(int port) {
        this.port = port;
    }
    
    void setLocalPort(int localport) {
        this.localport = localport;
    }
    
    @Override
    protected synchronized int available() throws IOException {
        if (this.isClosedOrPending()) {
            throw new IOException("Stream closed.");
        }
        if (this.isConnectionReset() || this.shut_rd) {
            return 0;
        }
        int n = 0;
        try {
            n = this.socketAvailable();
            if (n == 0 && this.isConnectionResetPending()) {
                this.setConnectionReset();
            }
        }
        catch (ConnectionResetException ex) {
            this.setConnectionResetPending();
            try {
                n = this.socketAvailable();
                if (n == 0) {
                    this.setConnectionReset();
                }
            }
            catch (ConnectionResetException ex2) {}
        }
        return n;
    }
    
    @Override
    protected void close() throws IOException {
        synchronized (this.fdLock) {
            if (this.fd != null) {
                if (!this.stream) {
                    ResourceManager.afterUdpClose();
                }
                if (this.fdUseCount == 0) {
                    if (this.closePending) {
                        return;
                    }
                    this.closePending = true;
                    try {
                        this.socketPreClose();
                    }
                    finally {
                        this.socketClose();
                    }
                    this.fd = null;
                    return;
                }
                else if (!this.closePending) {
                    this.closePending = true;
                    --this.fdUseCount;
                    this.socketPreClose();
                }
            }
        }
    }
    
    @Override
    void reset() throws IOException {
        if (this.fd != null) {
            this.socketClose();
        }
        this.fd = null;
        super.reset();
    }
    
    @Override
    protected void shutdownInput() throws IOException {
        if (this.fd != null) {
            this.socketShutdown(0);
            if (this.socketInputStream != null) {
                this.socketInputStream.setEOF(true);
            }
            this.shut_rd = true;
        }
    }
    
    @Override
    protected void shutdownOutput() throws IOException {
        if (this.fd != null) {
            this.socketShutdown(1);
            this.shut_wr = true;
        }
    }
    
    @Override
    protected boolean supportsUrgentData() {
        return true;
    }
    
    @Override
    protected void sendUrgentData(int n) throws IOException {
        if (this.fd == null) {
            throw new IOException("Socket Closed");
        }
        this.socketSendUrgentData(n);
        return;
    }
    
    @Override
    protected void finalize() throws IOException {
        this.close();
    }
    
    FileDescriptor acquireFD() {
        synchronized (this.fdLock) {
            ++this.fdUseCount;
            return this.fd;
        }
    }
    
    void releaseFD() {
        synchronized (this.fdLock) {
            --this.fdUseCount;
            if (this.fdUseCount == -1 && this.fd != null) {
                try {
                    this.socketClose();
                }
                catch (IOException ex) {}
                finally {
                    this.fd = null;
                }
            }
        }
    }
    
    public boolean isConnectionReset() {
        synchronized (this.resetLock) {
            return this.resetState == this.CONNECTION_RESET;
        }
    }
    
    public boolean isConnectionResetPending() {
        synchronized (this.resetLock) {
            return this.resetState == this.CONNECTION_RESET_PENDING;
        }
    }
    
    public void setConnectionReset() {
        synchronized (this.resetLock) {
            this.resetState = this.CONNECTION_RESET;
        }
    }
    
    public void setConnectionResetPending() {
        synchronized (this.resetLock) {
            if (this.resetState == this.CONNECTION_NOT_RESET) {
                this.resetState = this.CONNECTION_RESET_PENDING;
            }
        }
    }
    
    public boolean isClosedOrPending() {
        synchronized (this.fdLock) {
            if (this.closePending || this.fd == null) {
                return true;
            }
            return false;
        }
    }
    
    public int getTimeout() {
        return this.timeout;
    }
    
    private void socketPreClose() throws IOException {
        this.socketClose0(true);
    }
    
    protected void socketClose() throws IOException {
        this.socketClose0(false);
    }
    
    abstract void socketCreate(boolean p0) throws IOException;
    
    abstract void socketConnect(InetAddress p0, int p1, int p2) throws IOException;
    
    abstract void socketBind(InetAddress p0, int p1) throws IOException;
    
    abstract void socketListen(int p0) throws IOException;
    
    abstract void socketAccept(SocketImpl p0) throws IOException;
    
    abstract int socketAvailable() throws IOException;
    
    abstract void socketClose0(boolean p0) throws IOException;
    
    abstract void socketShutdown(int p0) throws IOException;
    
    abstract void socketSetOption(int p0, boolean p1, Object p2) throws SocketException;
    
    abstract int socketGetOption(int p0, Object p1) throws SocketException;
    
    abstract void socketSendUrgentData(int p0) throws IOException;
    
    static {
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                System.loadLibrary("net");
                return null;
            }
        });
        extendedOptions = ExtendedSocketOptions.getInstance();
        clientSocketOptions = clientSocketOptions();
        serverSocketOptions = serverSocketOptions();
    }
}
